@@ -607,18 +607,19 @@ sna_tiling_fill_boxes(struct sna *sna,
 	for (tile.extents.y1 = tile.extents.y2 = region.extents.y1;
 	     tile.extents.y2 < region.extents.y2;
 	     tile.extents.y1 = tile.extents.y2) {
-		tile.extents.y2 = tile.extents.y1 + step;
-		if (tile.extents.y2 > region.extents.y2)
-			tile.extents.y2 = region.extents.y2;
+		int y2 = tile.extents.y1 + step;
+		if (y2 > region.extents.y2)
+			y2 = region.extents.y2;
+		tile.extents.y2 = y2;
 
 		for (tile.extents.x1 = tile.extents.x2 = region.extents.x1;
 		     tile.extents.x2 < region.extents.x2;
 		     tile.extents.x1 = tile.extents.x2) {
 			PixmapRec tmp;
-
-			tile.extents.x2 = tile.extents.x1 + step;
-			if (tile.extents.x2 > region.extents.x2)
-				tile.extents.x2 = region.extents.x2;
+			int x2 = tile.extents.x1 + step;
+			if (x2 > region.extents.x2)
+				x2 = region.extents.x2;
+			tile.extents.x2 = x2;
 
 			tile.data = NULL;
 
@@ -733,18 +734,19 @@ bool sna_tiling_blt_copy_boxes(struct sna *sna, uint8_t alu,
 	for (tile.extents.y1 = tile.extents.y2 = region.extents.y1;
 	     tile.extents.y2 < region.extents.y2;
 	     tile.extents.y1 = tile.extents.y2) {
-		tile.extents.y2 = tile.extents.y1 + step;
-		if (tile.extents.y2 > region.extents.y2)
-			tile.extents.y2 = region.extents.y2;
+		int y2 = tile.extents.y1 + step;
+		if (y2 > region.extents.y2)
+			y2 = region.extents.y2;
+		tile.extents.y2 = y2;
 
 		for (tile.extents.x1 = tile.extents.x2 = region.extents.x1;
 		     tile.extents.x2 < region.extents.x2;
 		     tile.extents.x1 = tile.extents.x2) {
 			int w, h;
-
-			tile.extents.x2 = tile.extents.x1 + step;
-			if (tile.extents.x2 > region.extents.x2)
-				tile.extents.x2 = region.extents.x2;
+			int x2 = tile.extents.x1 + step;
+			if (x2 > region.extents.x2)
+				x2 = region.extents.x2;
+			tile.extents.x2 = x2;
 
 			tile.data = NULL;
 
@@ -795,21 +797,6 @@ done:
 	return ret;
 }
 
-static bool
-box_intersect(BoxPtr a, const BoxRec *b)
-{
-	if (a->x1 < b->x1)
-		a->x1 = b->x1;
-	if (a->x2 > b->x2)
-		a->x2 = b->x2;
-	if (a->y1 < b->y1)
-		a->y1 = b->y1;
-	if (a->y2 > b->y2)
-		a->y2 = b->y2;
-
-	return a->x1 < a->x2 && a->y1 < a->y2;
-}
-
 bool
 sna_tiling_copy_boxes(struct sna *sna, uint8_t alu,
 		      PixmapPtr src, struct kgem_bo *src_bo, int16_t src_dx, int16_t src_dy,
@@ -819,32 +806,40 @@ sna_tiling_copy_boxes(struct sna *sna, uint8_t alu,
 	BoxRec extents, tile, stack[64], *clipped, *c;
 	PixmapRec p;
 	int i, step, tiling;
+	bool create = true;
 	bool ret = false;
 
 	extents = box[0];
 	for (i = 1; i < n; i++) {
-		if (extents.x1 < box[i].x1)
+		if (box[i].x1 < extents.x1)
 			extents.x1 = box[i].x1;
-		if (extents.y1 < box[i].y1)
+		if (box[i].y1 < extents.y1)
 			extents.y1 = box[i].y1;
 
-		if (extents.x2 > box[i].x2)
+		if (box[i].x2 > extents.x2)
 			extents.x2 = box[i].x2;
-		if (extents.y2 > box[i].y2)
+		if (box[i].y2 > extents.y2)
 			extents.y2 = box[i].y2;
 	}
 
-	step = sna->render.max_3d_size - 4096 / dst->drawable.bitsPerPixel;
-	while (step * step * 4 > sna->kgem.max_upload_tile_size)
-		step /= 2;
-
 	tiling = I915_TILING_X;
 	if (!kgem_bo_can_blt(&sna->kgem, src_bo) ||
 	    !kgem_bo_can_blt(&sna->kgem, dst_bo))
 		tiling = I915_TILING_Y;
 
-	DBG(("%s: tiling copy, using %dx%d %c tiles\n",
-	     __FUNCTION__, step, step, tiling == I915_TILING_X ? 'X' : 'Y'));
+	create = (src_bo->pitch > sna->render.max_3d_pitch ||
+		  dst_bo->pitch > sna->render.max_3d_pitch);
+
+	step = sna->render.max_3d_size / 2;
+	if (create) {
+		while (step * step * 4 > sna->kgem.max_upload_tile_size)
+			step /= 2;
+	}
+
+	DBG(("%s: tiling copy %dx%d, %s %dx%d %c tiles\n", __FUNCTION__,
+	     extents.x2-extents.x1, extents.y2-extents.y1,
+	     create ? "creating" : "using",
+	     step, step, tiling == I915_TILING_X ? 'X' : 'Y'));
 
 	if (n > ARRAY_SIZE(stack)) {
 		clipped = malloc(sizeof(BoxRec) * n);
@@ -858,16 +853,17 @@ sna_tiling_copy_boxes(struct sna *sna, uint8_t alu,
 	p.devPrivate.ptr = NULL;
 
 	for (tile.y1 = extents.y1; tile.y1 < extents.y2; tile.y1 = tile.y2) {
-		tile.y2 = tile.y1 + step;
-		if (tile.y2 > extents.y2)
-			tile.y2 = extents.y2;
+		int y2 = tile.y1 + step;
+		if (y2 > extents.y2)
+			y2 = extents.y2;
+		tile.y2 = y2;
 
 		for (tile.x1 = extents.x1; tile.x1 < extents.x2; tile.x1 = tile.x2) {
 			struct kgem_bo *tmp_bo;
-
-			tile.x2 = tile.x1 + step;
-			if (tile.x2 > extents.x2)
-				tile.x2 = extents.x2;
+			int x2 = tile.x1 + step;
+			if (x2 > extents.x2)
+				x2 = extents.x2;
+			tile.x2 = x2;
 
 			c = clipped;
 			for (i = 0; i < n; i++) {
@@ -893,24 +889,31 @@ sna_tiling_copy_boxes(struct sna *sna, uint8_t alu,
 			DBG(("%s: tile (%d, %d), (%d, %d)\n",
 			     __FUNCTION__, tile.x1, tile.y1, tile.x2, tile.y2));
 
-			tmp_bo = kgem_create_2d(&sna->kgem,
-						p.drawable.width,
-						p.drawable.height,
-						p.drawable.bitsPerPixel,
-						tiling, CREATE_TEMPORARY);
-			if (!tmp_bo)
-				goto tiled_error;
+			if (create) {
+				tmp_bo = kgem_create_2d(&sna->kgem,
+							p.drawable.width,
+							p.drawable.height,
+							p.drawable.bitsPerPixel,
+							tiling, CREATE_TEMPORARY);
+				if (!tmp_bo)
+					goto tiled_error;
 
-			i = (sna->render.copy_boxes(sna, GXcopy,
-						    src, src_bo, src_dx, src_dy,
-						    &p, tmp_bo, -tile.x1, -tile.y1,
-						    clipped, c - clipped, 0) &&
-			     sna->render.copy_boxes(sna, alu,
-						    &p, tmp_bo, -tile.x1, -tile.y1,
-						    dst, dst_bo, dst_dx, dst_dy,
-						    clipped, c - clipped, 0));
+				i = (sna->render.copy_boxes(sna, GXcopy,
+							    src, src_bo, src_dx, src_dy,
+							    &p, tmp_bo, -tile.x1, -tile.y1,
+							    clipped, c - clipped, 0) &&
+				     sna->render.copy_boxes(sna, alu,
+							    &p, tmp_bo, -tile.x1, -tile.y1,
+							    dst, dst_bo, dst_dx, dst_dy,
+							    clipped, c - clipped, 0));
 
-			kgem_bo_destroy(&sna->kgem, tmp_bo);
+				kgem_bo_destroy(&sna->kgem, tmp_bo);
+			} else {
+				i = sna->render.copy_boxes(sna, GXcopy,
+							   src, src_bo, src_dx, src_dy,
+							   dst, dst_bo, dst_dx, dst_dy,
+							   clipped, c - clipped, 0);
+			}
 
 			if (!i)
 				goto tiled_error;
